#!/usr/bin/python3
#-*- coding: iso-8859-15 -*-

#############################################################################
#                                                                           #
#                               DS3231.py                                   #
#                                                                           #
# Mdulo para controlar el chip DS3231 con distintas funciones de usuario.  #
#                                                                           #
#############################################################################
#                                                                           #
# Autor: Alberto Martn Montero                                             #
# Fecha de la ltima revisin: 28/02/2018                                   #
#                                                                           #
# Trabajo Fin de Grado                                                      #
# Escuela de Ingenieras Industriales                                       #
#                                                                           #
#############################################################################

import RPi.GPIO as GPIO
from datetime import datetime
from smbus import SMBus
import time

# Definicin del siglo actual
SIGLO = 21

# Transforma un nmero binario (BCD 2x4bits) en uno entero
def bcd_to_int(bcd):
    n = 0
    for i in (bcd >> 4, bcd):
        for j in (1, 2, 4, 8):
            if i & 1:
                n += j
            i >>= 1
        n *= 10
    return n / 10

# Transforma un nmero entero en uno binario (BCD 2x4bits)
def int_to_bcd(n):
    bcd = 0
    for i in (n // 10, n % 10):
        for j in (8, 4, 2, 1):
            if i >= j:
                bcd += 1
                i -= j
            bcd <<= 1	
    return bcd >> 1

# Clase que contiene cada uno de los registros  y funciones del DS3231
class DS3231():
# Definicin de los registros
    REG_SEGUNDOS = 0x00
    REG_MINUTOS = 0x01
    REG_HORA = 0x02
    REG_DIA_SEMANA = 0x03
    REG_DIA = 0x04
    REG_MES = 0x05
    REG_ANIO = 0x06
    REG_A1_SEGUNDOS = 0x07
    REG_A1_MINUTOS = 0x08
    REG_A1_HORA = 0x09
    REG_A1_DIA = 0x0a
    REG_A2_MINUTOS = 0x0b
    REG_A2_HORA = 0x0c
    REG_A2_DIA = 0x0d
    REG_CONTROL = 0x0e
    REG_STATUS = 0x0f
    REG_TEMP_MSB = 0x11
    REG_TEMP_LSB = 0x12

#############################################################################
#                                                                           #
#                                   INIT                                    #
#                                                                           #
#############################################################################
    
    def __init__(self, direccion=0x68):
        #Detecta automticamente la versin de la RPi para crear el bus I2C
        if GPIO.RPI_REVISION == 1:
            i2c_bus = 0
        elif GPIO.RPI_REVISION == 2:
            i2c_bus = 1
        elif GPIO.RPI_REVISION == 3:
            i2c_bus = 1
        else:
            raise ValueError("No ha sido posible determinar la version " +
                             "de la Raspberry Pi" +
                             "\nNo ha sido posible establecer la " +
                             "comunicacion con el RTC")

        # print("Version de la Raspberry Pi: " + str(GPIO.RPI_REVISION))
        print("Creando bus I2C...")
        self.bus = SMBus(i2c_bus)
        self.direccion = direccion
        print("Bus I2C creado")

#############################################################################
#                                                                           #
#                                 ESCRIBIR                                  #
#                                                                           #
#############################################################################

    # Funcin primitiva para escritura de un dato en un registro
    def escribir(self, registro, dato):
        self.bus.write_byte_data(self.direccion, registro, dato)

    # Funciones especficas para la escritura en cada uno de los registros
    # Realizan una llamada a la funcin primitiva
    def escribir_fecha(self, hora, minutos, segundos, dia_semana, dia_mes,
                       mes, anio):
        if hora is not None:
            if hora < 0 or hora > 23:
                raise ValueError("Horas fuera de rango [0,23]")
            self.escribir(self.REG_HORA, int_to_bcd(hora))

        if minutos is not None:
            if minutos < 0 or minutos > 59:
                raise ValueError("Minutos fuera de rango [0,59]")
            self.escribir(self.REG_MINUTOS, int_to_bcd(minutos))

        if segundos is not None:
            if segundos < 0 or segundos > 59:
                raise ValueError("Segundos fuera de rango [0,59]")
            self.escribir(self.REG_SEGUNDOS, int_to_bcd(segundos))

        if dia_semana is not None:
            if dia_semana < 1 or dia_semana > 7:
                raise ValueError("Dia de la semana fuera de rango [1,7]")
            self.escribir(self.REG_DIA_SEMANA, int_to_bcd(dia_semana))

        if dia_mes is not None:
            if dia_mes < 1 or dia_mes > 31:
                raise ValueError("Dia del mes fuera de rango [1,31]")
            self.escribir(self.REG_DIA, int_to_bcd(dia_mes))

        if mes is not None:
            if mes < 1 or mes > 12:
                raise ValueError("Mes fuera de rango [1,12]")
            self.escribir(self.REG_MES, int_to_bcd(mes))

        if anio is not None:
            if anio < 0 or anio > 99:
                raise ValueError("Anio fuera de rango [0,99]")
            self.escribir(self.REG_ANIO, int_to_bcd(anio))

    def actualizar_fecha(self):
        dt = datetime.now()
        self.escribir_fecha(dt.hour, dt.minute, dt.second, dt.isoweekday(),
                           dt.day, dt.month, dt.year % 100)

#############################################################################
#                                                                           #
#                                   LEER                                    #
#                                                                           #
#############################################################################

    # Funcin primitiva para lectura de registro a travs del bus                
    def leer(self, registro):
        return self.bus.read_byte_data(self.direccion, registro)

    # Funciones especficas para la lectura de cada uno de los parmetros
    # Realizan una llamada a la funcin primitiva
    def leer_segundos(self):
        return bcd_to_int(self.leer(self.REG_SEGUNDOS))

    def leer_minutos(self):
        return bcd_to_int(self.leer(self.REG_MINUTOS))

    def leer_hora(self):
        return bcd_to_int(self.leer(self.REG_HORA))

    def leer_dia_semana(self):
        return bcd_to_int(self.leer(self.REG_DIA_SEMANA))

    def leer_dia_mes(self):
        return bcd_to_int(self.leer(self.REG_DIA))

    def leer_mes(self):
        return bcd_to_int(self.leer(self.REG_MES))

    def leer_anio(self):
        return bcd_to_int(self.leer(self.REG_ANIO))

    def leer_todo(self):
        return (self.leer_hora(), self.leer_minutos(), self.leer_segundos(),
                self.leer_dia_semana(), self.leer_dia_mes(), self.leer_mes(),
                self.leer_anio())

##    # Devuelve la lectura del registro en forma de string con formato
##    def leer_str(self):
##        return ("%02d:%02d:%02d - %02d/%02d/%02d" % (self.leer_hora(),
##                self.leer_minutos(), self.leer_segundos(), self.leer_dia_mes(),
##                self.leer_mes(), (SIGLO-1)*100+self.leer_anio()))

#############################################################################
#                                                                           #
#                                TEMPERATURA                                #
#                                                                           #
#############################################################################

    def obtener_temperatura(self):
        byte_msb = self.leer(self.REG_TEMP_MSB)
        byte_lsb = bin(self.leer(self.REG_TEMP_LSB))[2:].zfill(8)
        return byte_msb + int(byte_lsb[0])*2**(-1) + int(byte_lsb[1])*2**(-2)

#############################################################################
#                                                                           #
#                                  ALARMAS                                  #
#                                                                           #
#############################################################################

##### ALARMA 1
    
    def crear_alarma_1(self, modo, hora, minutos, dia=None):
        
        # MODO 1 - Alarma que se activa a la hora y minutos asignados
        #          (cualquier dia)
        if modo == 1:
            # Se definen los segundos a los que se activar la alarma (0)
            self.escribir(self.REG_A1_SEGUNDOS, 0x00)
            
            if hora < 0 or hora > 23:
                raise ValueError("Horas fuera de rango [0,23]")
            self.escribir(self.REG_A1_HORA, int_to_bcd(hora))

            if minutos < 0 or minutos > 59:
                raise ValueError("Minutos fuera de rango [0,59]")
            self.escribir(self.REG_A1_MINUTOS, int_to_bcd(minutos))
            
            # Se fija para que la alarma solo se active a la hora, minutos
            # y segundos especificados
            self.escribir(self.REG_A1_DIA, 0x80)
            
        # MODO 2 - Alarma que se activa a la hora, minutos y dia del mes
        #          asignados
        elif modo == 2:
            # Se definen los segundos a los que se activara la alarma (0)
            self.escribir(self.REG_A1_SEGUNDOS, 0x00)
            
            if hora < 0 or hora > 23:
                raise ValueError("Horas fuera de rango [0,23]")
            self.escribir(self.REG_A1_HORA, int_to_bcd(hora))

            if minutos < 0 or minutos > 59:
                raise ValueError("Minutos fuera de rango [0,59]")
            self.escribir(self.REG_A1_MINUTOS, int_to_bcd(minutos))
            
            if dia < 1 or dia > 31:
                raise ValueError("Dia del mes fuera de rango [1,31]")
            self.escribir(self.REG_A1_DIA, int_to_bcd(dia))
            
        # MODO 3 - Alarma que se activa a la hora, minutos y dia de la
        #          semana asignados
        elif modo == 3:
            # Se definen los segundos a los que se activara la alarma (0)
            self.escribir(self.REG_A1_SEGUNDOS, 0x00)
            
            if hora < 0 or hora > 23:
                raise ValueError("Horas fuera de rango [0,23]")
            self.escribir(self.REG_A1_HORA, int_to_bcd(hora))

            if minutos < 0 or minutos > 59:
                raise ValueError("Minutos fuera de rango [0,59]")
            self.escribir(self.REG_A1_MINUTOS, int_to_bcd(minutos))
            
            if dia < 1 or dia > 7:
                raise ValueError("Dia del mes fuera de rango [1,7]")
            self.escribir(self.REG_A1_DIA, int_to_bcd(dia)+0b1000000)
            # Bit 6 a 1 para habilitar dia de la
            # semana en vez de dia del mes
            
        # ERROR
        else:
            raise ValueError("Modo de alarma seleccionado erroneo")

    def activar_alarma_1(self):
        #Comprueba que en bit INTCN est a 1
        if (self.leer(self.REG_CONTROL) & 0b100):
            # Ya est a 1, no hace nada
            pass
        else:
            # Pone le bit a 1
            estado_registro = self.leer(self.REG_CONTROL)
            self.escribir(self.REG_CONTROL, estado_registro + 0b100)
        
        # Comprueba si ya est activa la alarma (bit 0 a 1)
        if (self.leer(self.REG_CONTROL) & 0b1):
            # Ya est activa, no hace nada
            pass
        else:
            # No est activa, escribe bit 0 a 1
            estado_registro = self.leer(self.REG_CONTROL)
            self.escribir(self.REG_CONTROL, estado_registro + 0b1)

    def desactivar_alarma_1(self):
        # Comprueba si ya est desactivada la alarma (bit 0 a 0)
        if ((self.leer(self.REG_CONTROL) & 0b1) == False):
            # Ya est desactiva, no hace nada
            pass
        else:
            # No est desactivada, escribe bit 0 a 0 (restndole 1)
            estado_registro = self.leer(self.REG_CONTROL)
            self.escribir(self.REG_CONTROL, estado_registro - 0b1)

    def leer_alarma_1(self):
        if (self.leer(self.REG_A1_DIA) & 0x80):
            modo = 1
            dia = None            
        elif ((self.leer(self.REG_A1_DIA) & 0x40) == False):
            modo = 2
            dia = bcd_to_int(self.leer(self.REG_A1_DIA))
        elif (self.leer(self.REG_A1_DIA) & 0x40):
            modo = 3
            dia = bcd_to_int(self.leer(self.REG_A1_DIA) - 0x40)
        else:
            raise ValueError("No se ha podido realizar lectura de " +
                             "la ALARMA 1")
        
        return (modo, bcd_to_int(self.leer(self.REG_A1_HORA)),
                bcd_to_int(self.leer(self.REG_A1_MINUTOS)), dia)

##### ALARMA 2
    
    def crear_alarma_2(self, modo, hora, minutos, dia=None):
        
        # MODO 1 - Alarma que se activa a la hora y minutos asignados
        #          (cualquier dia)
        if modo == 1:         
            if hora < 0 or hora > 23:
                raise ValueError("Horas fuera de rango [0,23]")
            self.escribir(self.REG_A2_HORA, int_to_bcd(hora))

            if minutos < 0 or minutos > 59:
                raise ValueError("Minutos fuera de rango [0,59]")
            self.escribir(self.REG_A2_MINUTOS, int_to_bcd(minutos))
            
            # Se fija para que la alarma solo se active a la hora, minutos
            # y segundos especificados
            self.escribir(self.REG_A2_DIA, 0x80)
            
        # MODO 2 - Alarma que se activa a la hora, minutos y dia del mes
        #          asignados
        elif modo == 2:
            if hora < 0 or hora > 23:
                raise ValueError("Horas fuera de rango [0,23]")
            self.escribir(self.REG_A2_HORA, int_to_bcd(hora))

            if minutos < 0 or minutos > 59:
                raise ValueError("Minutos fuera de rango [0,59]")
            self.escribir(self.REG_A2_MINUTOS, int_to_bcd(minutos))
            
            if dia < 1 or dia > 31:
                raise ValueError("Dia del mes fuera de rango [1,31]")
            self.escribir(self.REG_A2_DIA, int_to_bcd(dia))
            
        # MODO 3 - Alarma que se activa a la hora, minutos y dia de la
        #          semana asignados
        elif modo == 3:
            if hora < 0 or hora > 23:
                raise ValueError("Horas fuera de rango [0,23]")
            self.escribir(self.REG_A2_HORA, int_to_bcd(hora))

            if minutos < 0 or minutos > 59:
                raise ValueError("Minutos fuera de rango [0,59]")
            self.escribir(self.REG_A2_MINUTOS, int_to_bcd(minutos))
            
            if dia < 1 or dia > 7:
                raise ValueError("Dia del mes fuera de rango [1,7]")
            self.escribir(self.REG_A2_DIA, int_to_bcd(dia) + 0b1000000)
            # Bit 6 a 1 para habilitar dia de la
            # semana en vez de dia del mes
            
        # ERROR
        else:
            raise ValueError("Modo de alarma seleccionado erroneo")

    def activar_alarma_2(self):
        #Comprueba que en bit INTCN est a 1
        if (self.leer(self.REG_CONTROL) & 0b100):
            # Ya est a 1, no hace nada
            pass
        else:
            # Pone le bit a 1
            estado_registro = self.leer(self.REG_CONTROL)
            self.escribir(self.REG_CONTROL, estado_registro + 0b100)
        
        # Comprueba si ya est activa la alarma (bit 1 a 1)
        if (self.leer(self.REG_CONTROL) & 0b10):
            # Ya est activa, no hace nada
            pass
        else:
            # No est activa, escribe bit 1 a 1
            estado_registro = self.leer(self.REG_CONTROL)
            self.escribir(self.REG_CONTROL, estado_registro + 0b10)

    def desactivar_alarma_2(self):
        # Comprueba si ya est desactivada la alarma (bit 1 a 0)
        if ((self.leer(self.REG_CONTROL) & 0b10) == False):
            # Ya est desactiva, no hace nada
            pass
        else:
            # No est desactivada, escribe bit 1 a 0 (restndole 1)
            estado_registro = self.leer(self.REG_CONTROL)
            self.escribir(self.REG_CONTROL, estado_registro - 0b10)

    def leer_alarma_2(self):
        if (self.leer(self.REG_A2_DIA) & 0x80):
            modo = 1
            dia = None            
        elif ((self.leer(self.REG_A2_DIA) & 0x40) == False):
            modo = 2
            dia = bcd_to_int(self.leer(self.REG_A2_DIA))
        elif (self.leer(self.REG_A2_DIA) & 0x40):
            modo = 3
            dia = bcd_to_int(self.leer(self.REG_A2_DIA) - 0x40)
        else:
            raise ValueError("No se ha podido realizar lectura de " +
                             "la ALARMA 2")
        
        return (modo, bcd_to_int(self.leer(self.REG_A2_HORA)),
                bcd_to_int(self.leer(self.REG_A2_MINUTOS)), dia)

##### FLAGS

    def borrar_flags(self):
        # Flag alarma 1
        if (self.leer(self.REG_STATUS) & 0b1):
            estado_registro = self.leer(self.REG_STATUS)
            self.escribir(self.REG_STATUS, estado_registro - 0b1)

        #Flag alarma 2    
        if (self.leer(self.REG_STATUS) & 0b10):
            estado_registro = self.leer(self.REG_STATUS)
            self.escribir(self.REG_STATUS, estado_registro - 0b10)
